#!/usr/bin/env python
# -*- coding: utf-8 -*-

import Mesh
import MeshPart
import math
from math import log10, floor
import tempfile
import sys
import gettext
import wx
import wx.grid as gridlib
import glob
import os
import shutil
import re
import gettext

global doc, setEnvDefault
doc = App.ActiveDocument
setEnvDefault = ". /opt/OpenFOAM/OpenFOAM-2.3.x/etc/bashrc"

#---------------------------------------------------------------------------
class CustomDataTable(gridlib.PyGridTableBase):
    def __init__(self, log):
        gridlib.PyGridTableBase.__init__(self)
        self.log = log
        self.colLabels = ['objectName', 'Type', 'Refinement', 'cellSize', 'Boundary\nLayer', 'nLayer', 'ratio']
        self.dataTypes = [gridlib.GRID_VALUE_STRING,
                          gridlib.GRID_VALUE_CHOICE + ':empty,patch,wall,symmetryPlane,region',
                          gridlib.GRID_VALUE_CHOICE + ':local,object',
                          gridlib.GRID_VALUE_STRING,
                          gridlib.GRID_VALUE_BOOL,
                          gridlib.GRID_VALUE_NUMBER + ':0,9',
                          gridlib.GRID_VALUE_STRING
                          ]

        self.data =[]
        #--------------------------------------------------
        # FreeCAD上で表示されているすべてのobjについて、とりあえずデフォルト値を設定		
        #--------------------------------------------------
        for obj in doc.Objects:	
          if obj.ViewObject.Visibility:	# FreeCAD上で表示されているすべてのobj
             self.data = self.data + [[ obj.Label, "patch", "local", "", 0, 2, "1.2"]]
    #--------------------------------------------------
    # required methods for the wxPyGridTableBase interface
    def GetNumberRows(self):
        return len(self.data) + 1
    def GetNumberCols(self):
        return len(self.data[0])
    def IsEmptyCell(self, row, col):
        try:
            return not self.data[row][col]
        except IndexError:
            return True
    # Get/Set values in the table.  The Python version of these
    # methods can handle any data-type, (as long as the Editor and
    # Renderer understands the type too,) not just strings as in the
    # C++ version.
    def GetValue(self, row, col):
        try:
            return self.data[row][col]
        except IndexError:
            return ''
    def SetValue(self, row, col, value):
        try:
            self.data[row][col] = value
            #print row,col,self.data[row][col]
        except IndexError:
            # add a new row
            self.data.append([''] * self.GetNumberCols())
            self.SetValue(row, col, value)
            # tell the grid we've added a row
            msg = gridlib.GridTableMessage(self,            # The table
                    gridlib.GRIDTABLE_NOTIFY_ROWS_APPENDED, # what we did to it
                    1                                       # how many
                    )
            self.GetView().ProcessTableMessage(msg)
    #--------------------------------------------------
    # Some optional methods
    # Called when the grid needs to display labels
    def GetColLabelValue(self, col):
        return self.colLabels[col]
    # Called to determine the kind of editor/renderer to use by
    # default, doesn't necessarily have to be the same type used
    # natively by the editor/renderer if they know how to convert.
    def GetTypeName(self, row, col):
        return self.dataTypes[col]
    # Called to determine how the data can be fetched and stored by the
    # editor and renderer.  This allows you to enforce some type-safety
    # in the grid.
    def CanGetValueAs(self, row, col, typeName):
        colType = self.dataTypes[col].split(':')[0]
        if typeName == colType:
            return True
        else:
            return False
    def CanSetValueAs(self, row, col, typeName):
        return self.CanGetValueAs(row, col, typeName)
#---------------------------------------------------------------------------
class CustTableGrid(gridlib.Grid):
    def __init__(self, parent, log):
        gridlib.Grid.__init__(self, parent, -1)
        self.table = CustomDataTable(log)
        # The second parameter means that the grid is to take ownership of the
        # table and will destroy it when done.  Otherwise you would need to keep
        # a reference to it and call it's Destroy method later.
        self.SetTable(self.table, True)
        self.SetRowLabelSize(0)
        self.SetMargins(0,0)
        self.AutoSizeColumns(False)
        gridlib.EVT_GRID_CELL_LEFT_DCLICK(self, self.OnLeftDClick)
    # I do this because I don't like the default behaviour of not starting the
    # cell editor on double clicks, but only a second click.
    def OnLeftDClick(self, evt):
        if self.CanEnableCellControl():
            self.EnableCellEditControl()
#--------------------------------------------------------------------------

class objSetting(wx.Frame):

    def __init__(self, parent, log):
        global caseFile, maxCellSize, minCellSize, featureAngle	


        cellMax = searchBound2cellMax()
        wx.Frame.__init__(
            self, parent, -1, "cfMesh settings from FreeCAD Model(.fcstd)", size=(510,280)
            )

        fullPathFilename = doc.FileName
        folderName = fullPathFilename[0:fullPathFilename.rfind("/")]
        self.doc = doc
        p = wx.Panel(self, -1, style=0)
        bs = wx.BoxSizer(wx.VERTICAL)
        bs1 = wx.BoxSizer(wx.HORIZONTAL)
        bs2 = wx.BoxSizer(wx.HORIZONTAL)
        bs3 = wx.BoxSizer(wx.HORIZONTAL)

        b3 = wx.Button(p, -1, "case")
        self.grid = CustTableGrid(p, log)
        b1 = wx.Button(p, -1, "Export...")
        b2 = wx.Button(p, -1, "Exit")
        caseFile = wx.StaticText(p, wx.ID_ANY, folderName)
        caseFile.SetMinSize((500, 17))
        ts1 = wx.StaticText(p, wx.ID_ANY, "maxCellSize")
        maxCellSize = wx.TextCtrl(p, wx.ID_ANY, cellMax)
        ts2 = wx.StaticText(p, wx.ID_ANY, "minCellSize")
        minCellSize = wx.TextCtrl(p, wx.ID_ANY, "")
        ts3 = wx.StaticText(p, wx.ID_ANY, "featureAngle")
        featureAngle = wx.TextCtrl(p, wx.ID_ANY, "30")
        b1.SetDefault()
        b3.SetDefault()
        self.Bind(wx.EVT_BUTTON, self.exportSettingFiles, b1)
        self.Bind(wx.EVT_BUTTON, self.OnExitButton, b2)
        self.Bind(wx.EVT_BUTTON, self.selectCase, b3)
        b1.Bind(wx.EVT_SET_FOCUS, self.OnButtonFocus)

        bs.Add(bs1, 0, 0, 0)
        bs1.Add(b3)
        bs1.Add(caseFile)
        bs.Add(bs2, 0, 0, 0)
        bs2.Add(ts1)
        bs2.Add(maxCellSize)
        bs2.Add(ts2)
        bs2.Add(minCellSize)
        bs2.Add(ts3)
        bs2.Add(featureAngle)
        bs.Add(self.grid, 1, wx.GROW|wx.ALL, 5)
        bs.Add(bs3, 0, 0, 0)
        bs3.Add(b1)
        bs3.Add(b2)
        p.SetSizer(bs)

    #--------------------------------------------------------------------------------------------
    # case ボタンが押された場合の処理
    #--------------------------------------------------------------------------------------------
    def selectCase(self, evt):

        d = wx.DirDialog(self, "Choose a directory:",style=wx.DD_DEFAULT_STYLE | wx.DD_NEW_DIR_BUTTON)
   
        if d.ShowModal() == wx.ID_OK:
            workdir = d.GetPath()
            caseFile.Label = workdir
        dlg.Destroy()
        return

    #--------------------------------------------------------------------------------------------
    # Export ボタンが押された場合の処理
    #--------------------------------------------------------------------------------------------
    def exportSettingFiles(self, evt):
        caseFolder = caseFile.Label
        maxCellSizeValue = maxCellSize.Value
        minCellSizeValue = minCellSize.Value
        featureAngleValue = featureAngle.Value

        if os.path.exists(caseFolder):
          answer = self.YesNoDialog(u"cfMesh用の設定ファイルを作成しますか？")
          if answer == wx.ID_YES :

            ### step1 ### makeSTLfile ##############################
            fullPathFilename = os.path.splitext(doc.FileName)[0]
            fileName = fullPathFilename[fullPathFilename.rfind("/")+1:]
            self.makeStlFile(caseFolder, fileName)
            ### step2 ### convert stl to fms file ##############################
            #		using surfaceFeatureEdges command
            inputStl = caseFolder + "/" + fileName + ".stl"
            outputFms = caseFolder + "/" + fileName + ".fms"
            outputFmsTemp = caseFolder + "/" + fileName
            cmd = setEnvDefault + ";surfaceFeatureEdges -angle  " + featureAngle.Value + " "
            cmd = cmd + inputStl + " " + outputFms
            os.system(cmd)
            os.system("mv " + outputFms + " " + outputFmsTemp)

            answer = self.MessageDialog(u"cfMesh用のfmsファイルを作成しました。")
            ### step3 ### modify fms file ##############################
            self.modyfyFms(outputFmsTemp)
            ### step4 ### make meshDict ##############################
            self.makeMeshDict(caseFolder)
        return
    #--------------------------------------------------------------------------------------------
    # Typeがregion以外のオブジェクトをstlファイル出力
    #--------------------------------------------------------------------------------------------
    def makeStlFile(self, caseFolder, fileName):
        #fullPathFilename = os.path.splitext(doc.FileName)[0]
        #fileName = fullPathFilename[fullPathFilename.rfind("/")+1:]

        outputFile=open(caseFolder + "/"+fileName+'.stl','w')

        i=0
        while (self.grid.table.GetValue(i,0)):

            for obj in doc.Objects:
                if obj.Label == self.grid.table.GetValue(i,0):		#	表中のobjectNameがobj.Labelに一致したobj
                    if self.grid.table.GetValue(i,1) != "region":	#	Typeがregion以外のものについてstl出力
                        tmpFile = tempfile.mkstemp()
                        file = tmpFile[1]
                        obj.Shape.exportStl(file)
                        importFile = open(file,'r')
                        temp = importFile.readlines()
                        for line in temp:
                            if 'endsolid' in line:
                                outputFile.write('endsolid ' + obj.Label + '\n')
                            elif 'solid' in line:
                                outputFile.write('solid ' + obj.Label + '\n')
                            else:
                                outputFile.write(line)
                        importFile.close()
                        os.remove(file)
            i = i + 1
            
        outputFile.close()
        answer = self.MessageDialog(u"cfMesh用のstlファイルを作成しました。")
        return
    #--------------------------------------------------------------------------------------------
    # surfaceFeatureEdgesを使ってstl->fms変換した場合、
    # デフォルトのtype名はemptyになってしまうので、これを指定のtype名に変更する（sedを使用）
    #--------------------------------------------------------------------------------------------
    def modyfyFms(self, outputFmsTemp):
        outputFms = outputFmsTemp + ".fms"
        sedCmd = "sed "
        i = 0	# テーブルの行番号（オブジェクトの並び順）
        j = 0	# type=region以外(stl出力されるオブジェクト)の並び順
        while (self.grid.table.GetValue(i,0)):
            i = i + 1
            if self.grid.table.GetValue(i-1,1) != "region":
                j = j + 1
                sedCmd = sedCmd + "-e '" + str(2+3*j)+","+str(3+3*j)+"s/empty/"
                sedCmd = sedCmd + self.grid.table.GetValue(i-1,1)+"/' "

        sedCmd = sedCmd + outputFmsTemp + " > " + outputFms
        os.system(sedCmd)
        os.system("rm -f " + outputFmsTemp)

        answer = self.MessageDialog(u"cfMesh用fmsファイルのBoundaryTypeを修正しました。")
        return
    #--------------------------------------------------------------------------------------------
    #  objSettingのGUIパネルにて指定した各種パラメタを読み取って、meshDictを作成する
    #--------------------------------------------------------------------------------------------
    def makeMeshDict(self, caseFolder):

        meshDict = open(caseFolder + "/system/meshDict" , 'w')

        meshDict.write('FoamFile' + '\n')
        meshDict.write('{' + '\n')
        meshDict.write('version	2;' + '\n')
        meshDict.write('format	ascii;' + '\n')
        meshDict.write('class	dictionary;' + '\n')
        meshDict.write('location	"system";' + '\n')
        meshDict.write('object	meshDict;' + '\n')
        meshDict.write('}' + '\n')
        #meshDict.write('boundaryCellSize' + '\t' + str(maxCellSize.Value) + ';\n')
        meshDict.write('keepCellsIntersectingBoundary	0;' + '\n')
        meshDict.write('maxCellSize' + '\t' + str(maxCellSize.Value) + ';\n')

        if str(minCellSize.Value) != "":
            meshDict.write('minCellSize' + '\t' + str(minCellSize.Value) + ';\n')

        meshDict.write('removeGluedMesh	1;' + '\n')

        fullPathFilename = os.path.splitext(doc.FileName)[0]
        fileName = fullPathFilename[fullPathFilename.rfind("/")+1:]
        meshDict.write('surfaceFile' + '\t"' + fileName + '.fms";\n')

        meshDict.write('boundaryLayers' + '\n')
        meshDict.write('{' + '\n')
        meshDict.write('\tpatchBoundaryLayers' + '\n')
        meshDict.write('\t{' + '\n')

        # nLayers, thicknessRatio // BoundaryLayer チェックされたもの
        i=0
        while (self.grid.table.GetValue(i,0)):
            i = i + 1
            if self.grid.table.GetValue(i-1,4) == True:
                meshDict.write('\t\t' + str(self.grid.table.GetValue(i-1,0)) + '\n')
                meshDict.write('\t\t{' + '\n')
                meshDict.write('\t\t\tnLayers	' + str(self.grid.table.GetValue(i-1,5)) + ';\n')
                meshDict.write('\t\t\tthicknessRatio' + str(self.grid.table.GetValue(i-1,6)) + ';\n')
                meshDict.write('\t\t}' + '\n')
        # end of nLayers, thicknessRatio 

        meshDict.write('\t}' + '\n')
        meshDict.write('}' + '\n')
        meshDict.write('keepCellsIntersectingPatches' + '\n')
        meshDict.write('{' + '\n')
        meshDict.write('}' + '\n')
        meshDict.write('localRefinement' + '\n')
        meshDict.write('{' + '\n')

        # local refinement // cellSize指定されており、region以外のもの
        i=0
        while (self.grid.table.GetValue(i,0)):
            i = i + 1
            if (self.grid.table.GetValue(i-1,3) != "" and self.grid.table.GetValue(i-1,1) != "region"):
                meshDict.write('\t' + str(self.grid.table.GetValue(i-1,0)) + '\n')
                meshDict.write('\t{' + '\n')
                meshDict.write('\t\tcellSize\t' + str(self.grid.table.GetValue(i-1,3)) + ';\n')
                meshDict.write('\t}' + '\n')
        # end of local refinement // 

        meshDict.write('}' + '\n')
        meshDict.write('objectRefinements' + '\n')
        meshDict.write('{' + '\n')

        # object refinement // cellSize指定されており、Refinement=object
        # // Box要素であることを前提
        i=0
        boxNumber=0	
        while (self.grid.table.GetValue(i,0)):
            for obj in doc.Objects:
                if obj.Label == self.grid.table.GetValue(i,0):	
                    i = i + 1
                    if (str(self.grid.table.GetValue(i-1,3)) !="" and self.grid.table.GetValue(i-1,2) == "object"):
                        boxNumber = boxNumber + 1
                        xmax = obj.Shape.BoundBox.XMax
                        xmin = obj.Shape.BoundBox.XMin
                        ymax = obj.Shape.BoundBox.YMax
                        ymin = obj.Shape.BoundBox.YMin
                        zmax = obj.Shape.BoundBox.ZMax
                        zmin = obj.Shape.BoundBox.ZMin
                        centerX = 0.5*(xmax+xmin)
                        centerY = 0.5*(ymax+ymin)
                        centerZ = 0.5*(zmax+zmin)
                        meshDict.write('\tBox_#' + str(boxNumber) + '\n')
                        meshDict.write('\t{' + '\n')
                        meshDict.write('\t\tcellSize\t' + str(self.grid.table.GetValue(i-1,3)) + ';\n')
                        meshDict.write('\t\tcentre (' + str(centerX) + " " + str(centerY) + " " + str(centerZ) + ');\n')
                        meshDict.write('\t\tlengthX\t' + str(xmax-xmin) + ';\n')
                        meshDict.write('\t\tlengthY\t' + str(ymax-ymin) + ';\n')
                        meshDict.write('\t\tlengthZ\t' + str(zmax-zmin) + ';\n')
                        meshDict.write('\t\ttype box;' + '\n')
                        meshDict.write('\t}' + '\n')
        #end of object refinement // 

        meshDict.write('}' + '\n')

        meshDict.close()

        answer = self.MessageDialog(u"cfMesh用の設定ファイル(system/meshDict)を作成しました。")
        return
    #--------------------------------------------------------------------------------------------

    def OnButtonFocus(self, evt):
        pass

    ############################
    #  YesNoDialog
    #############################
    def YesNoDialog(self, message):
        dlg = wx.MessageDialog(self, message, "Question", wx.YES_NO)
        retCode = dlg.ShowModal()
        dlg.Destroy()
        return retCode
    ############################
    #  MessageDialog
    #############################
    def MessageDialog(self, message):
        dlg = wx.MessageDialog(self, message, "Message", wx.OK | wx.ICON_INFORMATION) 
        dlg.ShowModal()
        dlg.Destroy()


    def OnExitButton(self, evt):
        answer = self.YesNoDialog(u"終了しますか？")
        if answer == wx.ID_YES :
        	self.Close()
        return
#--------------------------------------------------------------------------------------------
# FreeCAD上で表示されているすべてのobjのBoundBox情報から、領域全体のBoundBox情報(DX,DY,DZ)を算出して、
# 概略セルサイズを決定する cellMax = ( DX + DY + DZ ) / 60
#--------------------------------------------------------------------------------------------
def searchBound2cellMax():
    xmax = -10000.0 
    xmin =  10000.0
    ymax = -10000.0 
    ymin =  10000.0
    zmax = -10000.0 
    zmin =  10000.0
    for obj in doc.Objects:
        if obj.ViewObject.Visibility:
            if obj.Shape.BoundBox.XMax > xmax:
                xmax = obj.Shape.BoundBox.XMax
            if obj.Shape.BoundBox.XMin < xmin:
                xmin = obj.Shape.BoundBox.XMin
            if obj.Shape.BoundBox.YMax > ymax:
                ymax = obj.Shape.BoundBox.YMax
            if obj.Shape.BoundBox.YMin < ymin:
                ymin = obj.Shape.BoundBox.YMin
            if obj.Shape.BoundBox.ZMax > zmax:
                zmax = obj.Shape.BoundBox.ZMax
            if obj.Shape.BoundBox.ZMin < zmin:
                zmin = obj.Shape.BoundBox.ZMin
    cellMax = round_sig((xmax-xmin+ymax-ymin+zmax-zmin)/60)
    cellMax = str(cellMax)
    
    return cellMax
#---------------------------------------------------------------------------
# 有効数字を2ケタに丸める
#---------------------------------------------------------------------------
def round_sig(x, sig=2):
    return round(x, sig-int(floor(log10(x)))-1)
#---------------------------------------------------------------------------
if __name__ == '__main__':

    app = wx.PySimpleApp(0)
    wx.InitAllImageHandlers()
    frame = objSetting(None, sys.stdout)
    frame.Show(True)
    app.MainLoop()
#---------------------------------------------------------------------------
